import decimal
import operator
import unittest

from prices import (FixedDiscount, History, LinearTax, Price, PriceRange,
                    inspect_price)


class PriceTest(unittest.TestCase):

    def setUp(self):
        self.ten_btc = Price(10, currency='BTC')
        self.twenty_btc = Price(20, currency='BTC')
        self.thirty_dollars = Price(30, currency='USD')

    def test_basics(self):
        self.assertEqual(self.ten_btc.net, self.ten_btc.gross)

    def test_subtraction(self):
        res = self.twenty_btc - self.ten_btc
        self.assertEqual(res, Price(10, currency='BTC'))

    def test_invalid_subtraction(self):
        self.assertRaises(ValueError,
                          lambda: self.ten_btc - self.thirty_dollars)
        self.assertRaises(TypeError,
                          lambda: self.ten_btc - 1)

    def test_multiplication(self):
        p1 = self.ten_btc * 5
        self.assertEqual(p1.net, 50)
        self.assertEqual(p1.gross, 50)
        p2 = self.ten_btc * 5
        self.assertEqual(p1, p2)

    def test_equality(self):
        p1 = Price(net='10', gross='20', currency='USD')
        p2 = Price(net='10', gross='20', currency='USD')
        p3 = Price(net='20', gross='20', currency='USD')
        p4 = Price(net='10', gross='10', currency='USD')
        p5 = Price(net='10', gross='20', currency='AUD')
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)
        self.assertNotEqual(p1, p4)
        self.assertNotEqual(p1, p5)
        self.assertNotEqual(p1, 10)

    def test_comparison(self):
        self.assertTrue(self.ten_btc < self.twenty_btc)
        self.assertTrue(self.twenty_btc > self.ten_btc)

    def test_invalid_comparison(self):
        self.assertRaises(ValueError,
                          lambda: self.ten_btc < self.thirty_dollars)

    def test_addition(self):
        p = self.ten_btc + self.twenty_btc
        self.assertEqual(p.net, 30)
        self.assertEqual(p.gross, 30)

    def test_invalid_addition(self):
        self.assertRaises(ValueError,
                          lambda: self.ten_btc + self.thirty_dollars)
        self.assertRaises(TypeError,
                          lambda: self.ten_btc + 1)

    def test_tax(self):
        p = Price(net='20', gross='30', currency='BTC')
        self.assertEqual(p.tax, 10)

    def test_inspect(self):
        p = ((self.ten_btc + self.twenty_btc) * 5 - self.ten_btc).quantize('0.01')
        self.assertEqual(
            inspect_price(p),
            "((((Price('10', currency='BTC') + Price('20', currency='BTC')) * 5) - Price('10', currency='BTC'))).quantize(Decimal('0.01'))")

    def test_elements(self):
        p1 = ((self.ten_btc + self.twenty_btc) * 5).quantize('0.01')
        self.assertEqual(
            list(p1.elements()),
            [self.ten_btc, self.twenty_btc, 5, decimal.Decimal('0.01')])
        p2 = Price(3, history=History(1, operator.__add__, 2))
        self.assertEqual(list(p2.elements()), [1, 2])

    def test_repr(self):
        p = Price(net='10', gross='20', currency='GBP')
        self.assertEqual(repr(p), "Price(net='10', gross='20', currency='GBP')")


class PriceRangeTest(unittest.TestCase):

    def setUp(self):
        self.ten_btc = Price(10, currency='BTC')
        self.twenty_btc = Price(20, currency='BTC')
        self.thirty_btc = Price(30, currency='BTC')
        self.forty_btc = Price(40, currency='BTC')
        self.range_ten_twenty = PriceRange(self.ten_btc, self.twenty_btc)
        self.range_thirty_forty = PriceRange(self.thirty_btc, self.forty_btc)

    def test_basics(self):
        self.assertEqual(self.range_ten_twenty.min_price, self.ten_btc)
        self.assertEqual(self.range_ten_twenty.max_price, self.twenty_btc)

    def test_construction(self):
        pr = PriceRange(self.ten_btc)
        self.assertEqual(pr.min_price, self.ten_btc)
        self.assertEqual(pr.max_price, self.ten_btc)

    def test_invalid_construction(self):
        p = Price(10, currency='USD')
        self.assertRaises(ValueError, lambda: PriceRange(self.ten_btc, p))
        self.assertRaises(ValueError, lambda: PriceRange(self.twenty_btc,
                                                         self.ten_btc))

    def test_addition(self):
        pr1 = self.range_ten_twenty + self.range_thirty_forty
        self.assertEqual(pr1.min_price, self.ten_btc + self.thirty_btc)
        self.assertEqual(pr1.max_price, self.twenty_btc + self.forty_btc)
        pr2 = self.range_ten_twenty + self.ten_btc
        self.assertEqual(pr2.min_price, self.ten_btc + self.ten_btc)
        self.assertEqual(pr2.max_price, self.twenty_btc + self.ten_btc)

    def test_invalid_addition(self):
        p = Price(10, currency='USD')
        pr = PriceRange(p)
        self.assertRaises(ValueError, lambda: self.range_ten_twenty + pr)
        self.assertRaises(ValueError, lambda: self.range_ten_twenty + p)
        self.assertRaises(TypeError, lambda: self.range_ten_twenty + 10)

    def test_subtraction(self):
        pr1 = self.range_thirty_forty - self.range_ten_twenty
        pr2 = self.range_thirty_forty - self.ten_btc
        self.assertEqual(pr1.min_price, self.thirty_btc - self.ten_btc)
        self.assertEqual(pr1.max_price, self.forty_btc - self.twenty_btc)
        self.assertEqual(pr2.min_price, self.thirty_btc - self.ten_btc)
        self.assertEqual(pr2.max_price, self.forty_btc - self.ten_btc)

    def test_invalid_subtraction(self):
        p = Price(10, currency='USD')
        pr = PriceRange(p)
        self.assertRaises(ValueError, lambda: self.range_thirty_forty - pr)
        self.assertRaises(ValueError, lambda: self.range_thirty_forty - p)
        self.assertRaises(TypeError, lambda: self.range_thirty_forty - 1)

    def test_equality(self):
        pr1 = PriceRange(self.ten_btc, self.twenty_btc)
        pr2 = PriceRange(self.ten_btc, self.twenty_btc)
        pr3 = PriceRange(self.ten_btc, self.ten_btc)
        pr4 = PriceRange(self.twenty_btc, self.twenty_btc)
        self.assertEqual(pr1, pr2)
        self.assertNotEqual(pr1, pr3)
        self.assertNotEqual(pr1, pr4)
        self.assertNotEqual(pr1, self.ten_btc)

    def test_membership(self):
        self.assertTrue(self.ten_btc in self.range_ten_twenty)
        self.assertTrue(self.twenty_btc in self.range_ten_twenty)
        self.assertFalse(self.thirty_btc in self.range_ten_twenty)

    def test_invalid_membership(self):
        self.assertRaises(TypeError, lambda: 15 in self.range_ten_twenty)

    def test_replacement(self):
        pr1 = self.range_ten_twenty.replace(max_price=self.thirty_btc)
        self.assertEqual(pr1.min_price, self.ten_btc)
        self.assertEqual(pr1.max_price, self.thirty_btc)
        pr2 = self.range_thirty_forty.replace(min_price=self.twenty_btc)
        self.assertEqual(pr2.min_price, self.twenty_btc)
        self.assertEqual(pr2.max_price, self.forty_btc)

    def test_repr(self):
        pr1 = self.range_thirty_forty
        pr2 = PriceRange(self.ten_btc, self.ten_btc)
        self.assertEqual(
            repr(pr1),
            "PriceRange(Price('30', currency='BTC'), Price('40', currency='BTC'))")
        self.assertEqual(
            repr(pr2),
            "PriceRange(Price('10', currency='BTC'))")


class LinearTaxTest(unittest.TestCase):

    def setUp(self):
        self.ten_btc = Price(10, currency='BTC')
        self.twenty_btc = Price(20, currency='BTC')
        self.range_ten_twenty = PriceRange(self.ten_btc, self.twenty_btc)

    def test_price(self):
        tax = LinearTax(1, name='2x Tax')
        p = self.ten_btc + tax
        self.assertEqual(p.net, self.ten_btc.net)
        self.assertEqual(p.gross, self.ten_btc.gross * 2)
        self.assertEqual(p.currency, self.ten_btc.currency)

    def test_pricerange(self):
        tax_name = '2x Tax'
        tax = LinearTax(1, name=tax_name)
        pr = self.range_ten_twenty + tax
        self.assertEqual(pr.min_price.net, self.ten_btc.net)
        self.assertEqual(pr.min_price.gross, self.ten_btc.gross * 2)
        self.assertEqual(pr.min_price.currency, self.ten_btc.currency)
        self.assertEqual(pr.max_price.net, self.twenty_btc.net)
        self.assertEqual(pr.max_price.gross, self.twenty_btc.gross * 2)
        self.assertEqual(pr.max_price.currency, self.twenty_btc.currency)

    def test_comparison(self):
        tax1 = LinearTax(1)
        tax2 = LinearTax(2)
        self.assertTrue(tax1 < tax2)
        self.assertTrue(tax2 > tax1)
        self.assertRaises(TypeError, lambda: tax1 < 10)

    def test_equality(self):
        tax1 = LinearTax(1)
        tax2 = LinearTax(1)
        tax3 = LinearTax(2)
        self.assertEqual(tax1, tax2)
        self.assertNotEqual(tax1, tax3)
        self.assertNotEqual(tax1, 1)

    def test_repr(self):
        tax = LinearTax(decimal.Decimal('1.23'), name='VAT')
        self.assertEqual(repr(tax), "LinearTax('1.23', name='VAT')")


class FixedDiscountTest(unittest.TestCase):

    def setUp(self):
        self.ten_btc = Price(10, currency='BTC')
        self.ten_usd = Price(10, currency='USD')
        self.thirty_btc = Price(30, currency='BTC')

    def test_discount(self):
        discount = FixedDiscount(self.ten_btc, name='Ten off')
        p = self.thirty_btc + discount
        self.assertEqual(p.net, 20)
        self.assertEqual(p.gross, 20)
        self.assertEqual(p.currency, 'BTC')

    def test_currency_mismatch(self):
        discount = FixedDiscount(self.ten_usd)
        self.assertRaises(ValueError, lambda: self.ten_btc + discount)

    def test_repr(self):
        discount = FixedDiscount(self.ten_usd, name='Ten off')
        self.assertEqual(
            repr(discount),
            "FixedDiscount(Price('10', currency='USD'), name='Ten off')")


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
import prices

from setuptools import setup

CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.5',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

setup(name='prices',
      author='Mirumee Software',
      author_email='hello@mirumee.com',
      description='Python price handling for humans',
      license='BSD',
      version='0.5',
      url='http://satchless.com/',
      packages=['prices'],
      test_suite='prices.tests',
      include_package_data=True,
      classifiers=CLASSIFIERS,
      platforms=['any'])



from collections import namedtuple
from decimal import Decimal, ROUND_HALF_UP
import operator
import warnings


class History(namedtuple('History', 'left operator right')):

    def __repr__(self):
        left = self.left.history if isinstance(self.left, Price) else None
        left = left or self.left
        right = self.right.history if isinstance(self.right, Price) else None
        right = right or self.right
        if self.operator is operator.__mul__:
            return '(%r * %r)' % (left, right)
        elif self.operator == Price.quantize:
            return '(%r).quantize(%r)' % (left, right)
        elif self.operator is operator.__sub__:
            return '(%r - %r)' % (left, right)
        return '(%r + %r)' % (left, right)


class Price(namedtuple('Price', 'net gross currency history')):

    def __new__(cls, net, gross=None, currency=None, history=None):
        if isinstance(net, float) or isinstance(gross, float):
            warnings.warn(  # pragma: no cover
                RuntimeWarning(
                    'You should never use floats when dealing with prices!'),
                stacklevel=2)
        net = Decimal(net)
        if gross is not None:
            gross = Decimal(gross)
        else:
            gross = net
        return super(Price, cls).__new__(cls, net, gross, currency, history)

    def __repr__(self):
        if self.net == self.gross:
            return 'Price(%r, currency=%r)' % (str(self.net), self.currency)
        return ('Price(net=%r, gross=%r, currency=%r)' %
                (str(self.net), str(self.gross), self.currency))

    def __lt__(self, other):
        if isinstance(other, Price):
            if self.currency != other.currency:
                raise ValueError('Cannot compare prices in %r and %r' %
                                 (self.currency, other.currency))
            return self.gross < other.gross
        return NotImplemented

    def __le__(self, other):
        return self < other or self == other

    def __eq__(self, other):
        if isinstance(other, Price):
            return (self.gross == other.gross and
                    self.net == other.net and
                    self.currency == other.currency)
        return False

    def __ne__(self, other):
        return not self == other

    def __mul__(self, other):
        price_net = self.net * other
        price_gross = self.gross * other
        history = History(self, operator.__mul__, other)
        return Price(net=price_net, gross=price_gross, currency=self.currency,
                     history=history)

    def __add__(self, other):
        if isinstance(other, PriceModifier):
            return other.apply(self)
        if isinstance(other, Price):
            if other.currency != self.currency:
                raise ValueError('Cannot add price in %r to %r' %
                                 (self.currency, other.currency))
            price_net = self.net + other.net
            price_gross = self.gross + other.gross
            history = History(self, operator.__add__, other)
            return Price(net=price_net, gross=price_gross,
                         currency=self.currency, history=history)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Price):
            if other.currency != self.currency:
                raise ValueError('Cannot subtract prices in %r from %r' %
                                 (other.currency, self.currency))
            price_net = self.net - other.net
            price_gross = self.gross - other.gross
            history = History(self, operator.__sub__, other)
            return Price(net=price_net, gross=price_gross,
                         currency=self.currency, history=history)
        return NotImplemented

    @property
    def tax(self):
        return self.gross - self.net

    def quantize(self, exp, rounding=ROUND_HALF_UP):
        exp = Decimal(exp)
        history = History(self, Price.quantize, exp)
        return Price(net=self.net.quantize(exp, rounding=rounding),
                     gross=self.gross.quantize(exp, rounding=rounding),
                     currency=self.currency, history=history)

    def elements(self):
        if not self.history:
            yield self
        else:
            if hasattr(self.history.left, 'elements'):
                for el in self.history.left.elements():
                    yield el
            else:
                yield self.history.left
            if hasattr(self.history.right, 'elements'):
                for el in self.history.right.elements():
                    yield el
            else:
                yield self.history.right


class PriceRange(namedtuple('PriceRange', 'min_price max_price')):

    def __new__(cls, min_price, max_price=None):
        if max_price is None:
            max_price = min_price
        if min_price.currency != max_price.currency:
            raise ValueError('Cannot create a pricerange as %r and %r use'
                             ' different currencies' % (min_price, max_price))
        if min_price > max_price:
            raise ValueError('Cannot create a pricerange from %r to %r' %
                             (min_price, max_price))
        return super(PriceRange, cls).__new__(cls, min_price, max_price)

    def __repr__(self):
        if self.max_price == self.min_price:
            return 'PriceRange(%r)' % (self.min_price,)
        return 'PriceRange(%r, %r)' % (self.min_price, self.max_price)

    def __add__(self, other):
        if isinstance(other, PriceModifier):
            return PriceRange(min_price=other.apply(self.min_price),
                              max_price=other.apply(self.max_price))
        if isinstance(other, Price):
            if other.currency != self.min_price.currency:
                raise ValueError("Cannot add pricerange in %r to price in %r" %
                                 (self.min_price.currency, other.currency))
            min_price = self.min_price + other
            max_price = self.max_price + other
            return PriceRange(min_price=min_price, max_price=max_price)
        elif isinstance(other, PriceRange):
            if other.min_price.currency != self.min_price.currency:
                raise ValueError('Cannot add priceranges in %r and %r' %
                                 (self.min_price.currency,
                                  other.min_price.currency))
            min_price = self.min_price + other.min_price
            max_price = self.max_price + other.max_price
            return PriceRange(min_price=min_price, max_price=max_price)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Price):
            if other.currency != self.min_price.currency:
                raise ValueError("Cannot subtract price in %r from pricerange"
                                 " in %r" %
                                 (other.currency, self.min_price.currency))
            min_price = self.min_price - other
            max_price = self.max_price - other
            return PriceRange(min_price=min_price, max_price=max_price)
        elif isinstance(other, PriceRange):
            if other.min_price.currency != self.min_price.currency:
                raise ValueError('Cannot subtract pricerange in %r from %r' %
                                 (other.min_price.currency,
                                  self.min_price.currency))
            min_price = self.min_price - other.min_price
            max_price = self.max_price - other.max_price
            return PriceRange(min_price=min_price, max_price=max_price)
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, PriceRange):
            return (self.min_price == other.min_price and
                    self.max_price == other.max_price)
        return False

    def __ne__(self, other):
        return not self == other

    def __contains__(self, item):
        if not isinstance(item, Price):
            raise TypeError('in <pricerange> requires price as left operand,'
                            ' not %s' % (type(item),))
        return self.min_price <= item <= self.max_price

    def replace(self, min_price=None, max_price=None):
        '''
        Return a new pricerange object with one or more properties set to
        values passed to this method.
        '''
        if min_price is None:
            min_price = self.min_price
        if max_price is None:
            max_price = self.max_price
        return PriceRange(min_price=min_price, max_price=max_price)


class PriceModifier(object):

    name = None

    def apply(self, price):
        raise NotImplementedError()


class Tax(PriceModifier):
    '''
    A generic tax class, provided so all taxers have a common base.
    '''
    def apply(self, price_obj):
        history = History(price_obj, operator.__add__, self)
        return Price(net=price_obj.net,
                     gross=price_obj.gross + self.calculate_tax(price_obj),
                     currency=price_obj.currency, history=history)

    def calculate_tax(self, price_obj):
        raise NotImplementedError()


class LinearTax(Tax):
    '''
    Adds a certain fraction on top of the price.
    '''
    def __init__(self, multiplier, name=None):
        self.multiplier = Decimal(multiplier)
        self.name = name or self.name

    def __repr__(self):
        return 'LinearTax(%r, name=%r)' % (str(self.multiplier), self.name)

    def __lt__(self, other):
        if not isinstance(other, LinearTax):
            raise TypeError('Cannot compare lineartax to %r' % (other,))
        return self.multiplier < other.multiplier

    def __eq__(self, other):
        if isinstance(other, LinearTax):
            return (self.multiplier == other.multiplier and
                    self.name == other.name)
        return False

    def __ne__(self, other):
        return not self == other

    def calculate_tax(self, price_obj):
        return price_obj.gross * self.multiplier


class FixedDiscount(PriceModifier):
    '''
    Adds a fixed amount to the price.
    '''
    def __init__(self, amount, name=None):
        self.amount = amount
        self.name = name or self.name

    def __repr__(self):
        return 'FixedDiscount(%r, name=%r)' % (self.amount, self.name)

    def apply(self, price_obj):
        if price_obj.currency != self.amount.currency:
            raise ValueError('Cannot apply a discount in %r to a price in %r' %
                             (self.amount.currency, price_obj.currency))
        history = History(price_obj, operator.__add__, self)
        return Price(net=price_obj.net - self.amount.net,
                     gross=price_obj.gross - self.amount.gross,
                     currency=price_obj.currency, history=history)


def inspect_price(price_obj):
    return repr(price_obj.history or price_obj)
